## ACOPF formulation
## acopf-V.mod AMPL file
## Leo Liberti
## This AMPL model implements the arc formulation described in
##   ../../papers/acopf/formulations/formulations.pdf (190930)
##    it can integrate shunts, transformers or both
##   Formulation in real numbers (rather than complex),
##     decision variables: generated power, voltage
## 191031  derived from acopf.mod through replacement I=YV, S=VI^*
##         => S = V(YV)^* = V Y^* V^* = Y^* VV^*

#### user-configurable ####
param Inf := 1e30;
param Eps := 1e-6;
param Pi := 4*atan(1);
param myZero := 1e-9;

## sets

# max number of parallel branches in instance
param maxParBranches integer, >0, default 1;
set PB := 1..maxParBranches;

# buses
set B;

## lines (directed: edges correspond to antisymmetric arcs)
## 191026: can't deal with parallel arcs (eg case57)
# the given edges (b,a): the transformer is on b
set L0 within {B,B,PB};  
# set of all antiparallel arcs
set L default L0 union {(a,b,i) in {B,B,PB} : (b,a,i) in L0};
set L1 default L diff L0;

# set of generators at node
set G{b in B} default {};

## parameters

# bus type (2=generator, 3=reference)
param busType{B} integer;

# cost coefficients (minimization of power generation costs)
param Kcard integer, >= 0, default 2;
set K := 0..Kcard;
# initialization: only linear terms in P (quadratic in V)
param C{b in B, g in G[b], k in K} default if k == 1 then 1 else 0;

# real power demand at buses (there can be nodes with negative demands)
param SDR{B} default 0; 
# reactive power demand at buses (appears in MATPOWER documentation)
param SDC{B} default 0;

# power bounds at generators (real, imaginary)
param SLR{b in B, G[b]} default -Inf;
param SLC{b in B, G[b]} default -Inf;
param SUR{b in B, g in G[b]} >= SLR[b,g], default Inf;
param SUC{b in B, g in G[b]} >= SLC[b,g], default Inf;

# upper power magnitude bounds on links - symmetric
param SU{L} >= 0, default Inf; 

# voltage magnitude bounds at buses
param VL{B} default myZero; # must have VL>0 (bound on injected power)
param VU{b in B} >= VL[b], default Inf;
param Vm{B} default 1; # unused
param Va{B} default 0; # reference buses have VC fixed at Va

# shunt parameters at buses
param shR{B} default 0;  # MatPower's Gs
param shC{B} default 0;  # MatPower's Bs

# status of a branch
param status{L} default 1;

# Y matrix (Ohm's law in AC) data, only defined on given arcs in L0
param r{L0} default 0;
param x{L0} default 0;
param bb{L0} default 0;
param tau{L0} default 1;
param theta{L0} default 0;  # translated to radians by mpc2dat.py

# phase difference bounds (only across lines in L0)
#   translated to radians by mpc2dat.py ##191026: not using these
param pdLB{L0} default -Pi;
param pdUB{L0} default Pi;

# the 2x2 complex Y matrix appearing in Ohm's law for a line (b,a) in L0
#   (computation in .run file)
param YbbR{L0} default 0;
param YbbC{L0} default 0;
param YbaR{L0} default 0;
param YbaC{L0} default 0;
param YabR{L0} default 0;
param YabC{L0} default 0;
param YaaR{L0} default 0;
param YaaC{L0} default 0;

param Ybb2{(b,a,i) in L0} default YbbR[b,a,i]^2 + YbbC[b,a,i]^2;
param Yba2{(b,a,i) in L0} default YbaR[b,a,i]^2 + YbaC[b,a,i]^2;
param Yab2{(b,a,i) in L0} default YabR[b,a,i]^2 + YabC[b,a,i]^2;
param Yaa2{(b,a,i) in L0} default YaaR[b,a,i]^2 + YaaC[b,a,i]^2;
param YbbYbaR{(b,a,i) in L0} default YbbR[b,a,i]*YbaR[b,a,i] + YbbC[b,a,i]*YbaC[b,a,i];
param YbbYbaC{(b,a,i) in L0} default YbbR[b,a,i]*YbaC[b,a,i] - YbbC[b,a,i]*YbaR[b,a,i];
param YabYaaR{(b,a,i) in L0} default YabR[b,a,i]*YaaR[b,a,i] + YabC[b,a,i]*YaaC[b,a,i];
param YabYaaC{(b,a,i) in L0} default YabR[b,a,i]*YaaC[b,a,i] - YabC[b,a,i]*YaaR[b,a,i];


# starting point for voltage variables
param VRval{B} default 0;
param VCval{B} default 0;

## decision variables

# voltage (real, imaginary)
var VR{B};
var VC{B};

# power generation (real, imaginary)
var SgenR{b in B, g in G[b]} >= SLR[b,g], <= SUR[b,g];
var SgenC{b in B, g in G[b]} >= SLC[b,g], <= SUC[b,g];

# aux var V2 = |V|^2
var V2{b in B} >= VL[b]^2, <= VU[b]^2;
# aux var VbaR = VR[b]*VR[a]+VC[b]*VC[a]
var VbaR{b in B, a in B : (b,a,1) in L};
# aux var VbaC = VR[b]*VC[a]-VC[b]*VR[a]
var VbaC{b in B, a in B : (b,a,1) in L};

## objective function

# generation cost (WARNING: to remove quadratic power terms set C[g,2]=0)
minimize generation_costs: sum{b in B, g in G[b]} (C[b,g,2]*SgenR[b,g]^2 + C[b,g,1]*SgenR[b,g] + C[b,g,0]);

## constraints

# power flow, derived from replacing S_ba with V_b(Y_bb V_b - Y_ba V_a)^* 
subject to power_flowRb{b in B}:  # real
   SDR[b] + sum{(b,a,i) in L0} (YbbR[b,a,i]*V2[b] + YbaR[b,a,i]*VbaR[b,a] - YbaC[b,a,i]*VbaC[b,a]) + sum{(b,a,i) in L1} (YaaR[a,b,i]*V2[b] + YabR[a,b,i]*VbaR[b,a] - YabC[a,b,i]*VbaC[b,a]) = -shR[b]*V2[b] + sum{g in G[b]} SgenR[b,g];
subject to power_flowCb{b in B}:  # imaginary
   SDC[b] + sum{(b,a,i) in L0} (-YbbC[b,a,i]*V2[b] - YbaC[b,a,i]*VbaR[b,a] - YbaR[b,a,i]*VbaC[b,a]) + sum{(b,a,i) in L1} (-YaaC[a,b,i]*V2[b] - YabC[a,b,i]*VbaR[b,a] - YabR[a,b,i]*VbaC[b,a]) =  shC[b]*V2[b] + sum{g in G[b]} SgenC[b,g];

# power bound on lines b->a defined on |S| <= SU:
#   reformulate SU => |S| = |V||I| >= VL|I| --> |I| <= SU/VL
subject to power_bound_to1{(b,a,i) in L0 : SU[b,a,i]>0 and SU[b,a,i]<Inf}:
   Ybb2[b,a,i]*V2[b] + Yba2[b,a,i]*V2[a] + 2*(YbbYbaR[b,a,i]*VbaR[b,a] + YbbYbaC[b,a,i]*VbaC[b,a]) <= SU[b,a,i]^2 / VU[b]^2;
subject to power_bound_to2{(b,a,i) in L0 : SU[b,a,i]>0 and SU[b,a,i]<Inf}:
   Yab2[b,a,i]*V2[b] + Yaa2[b,a,i]*V2[a] + 2*(YabYaaR[b,a,i]*VbaR[b,a] + YabYaaC[b,a,i]*VbaC[b,a]) <= SU[b,a,i]^2 / VU[b]^2;

# definition of V2
subject to V2_def{b in B}: V2[b] = VR[b]^2 + VC[b]^2;
# definition of VbaR
subject to VbaR_def{b in B, a in B : (b,a,1) in L}:
  VbaR[b,a] = VR[b]*VR[a] + VC[b]*VC[a];
# definition of VbaC
subject to VbaC_def{b in B, a in B : (b,a,1) in L}:
  VbaC[b,a] = VR[b]*VC[a] - VC[b]*VR[a];

# bounds on phase difference
subject to phasediff1{(b,a,i) in L0}: -VbaC[b,a] <= tan(pdUB[b,a,i])*VbaR[b,a];
subject to phasediff2{(b,a,i) in L0}: -VbaC[b,a] >= tan(pdLB[b,a,i])*VbaR[b,a];
subject to phasediff3{(b,a,1) in L0}: VbaR[b,a] >= 0;

# reference bus: there had better be just one reference -- check in .run
subject to reference1{b in B : busType[b] == 3}: VC[b] = 0;
subject to reference2{b in B : busType[b] == 3}: VR[b] >= 0;

