## ACOPF formulation - polar variant of S,I,V formulation (acopf.mod)
## acopf_polar.mod AMPL file
## Leo Liberti

#### user-configurable ####
param Inf := 1e30;
param Eps := 1e-6;
param Pi := 4*atan(1);
param myZero := 1e-9;

## sets

# max number of parallel branches in instance
param maxParBranches integer, >0, default 1;
set PB := 1..maxParBranches;

# buses
set B;

## lines (directed: edges correspond to antisymmetric arcs)
## 191026: can't deal with parallel arcs (eg case57)
# the given edges (b,a): the transformer is on b
set L0 within {B,B,PB};  
# set of all antiparallel arcs
set L default L0 union {(a,b,i) in {B,B,PB} : (b,a,i) in L0}; 

# set of generators at node
set G{b in B} default {};

## parameters

# bus type (2=generator, 3=reference)
param busType{B} integer;

# cost coefficients (minimization of power generation costs)
param Kcard integer, >= 0, default 2;
set K := 0..Kcard;
# initialization: only linear terms in P (quadratic in V)
param C{b in B, g in G[b], k in K} default if k == 1 then 1 else 0;

# real power demand at buses (there can be nodes with negative demands)
param SDR{B} default 0; 
# reactive power demand at buses (appears in MATPOWER documentation)
param SDC{B} default 0;

# power bounds at generators (real, imaginary)
param SLR{b in B, G[b]} default -Inf;
param SLC{b in B, G[b]} default -Inf;
param SUR{b in B, g in G[b]} >= SLR[b,g], default Inf;
param SUC{b in B, g in G[b]} >= SLC[b,g], default Inf;

# upper power magnitude bounds on links - symmetric
param SU{L} >= 0, default Inf; 

# voltage magnitude bounds at buses
param VL{B} default 0; # can't have negative moduli
param VU{b in B} >= VL[b], default Inf;
param Vm{B} default 1; # unused
param Va{B} default 0; # reference buses have VC fixed at Va

# shunt parameters at buses
param shR{B} default 0;  # MatPower's Gs
param shC{B} default 0;  # MatPower's Bs

# status of a branch
param status{L} default 1;

# Y matrix (Ohm's law in AC) data, only defined on given arcs in L0
param r{L0} default 0;
param x{L0} default 0;
param bb{L0} default 0;
param tau{L0} default 1;
param theta{L0} default 0;  # translated to radians by mpc2dat.py

# phase difference bounds (only across lines in L0)
#   translated to radians by mpc2dat.py ##191026: not using these
param pdLB{L0} default -Pi;
param pdUB{L0} default Pi;

# the 2x2 complex Y matrix appearing in Ohm's law for a line (b,a) in L0
#   (computation in .run file)
param YbbR{L0} default 0;
param YbbC{L0} default 0;
param YbaR{L0} default 0;
param YbaC{L0} default 0;
param YabR{L0} default 0;
param YabC{L0} default 0;
param YaaR{L0} default 0;
param YaaC{L0} default 0;

# starting point for voltage variables
param VRval{B} default 0;
param VCval{B} default 0;

## decision variables

# voltage (real, imaginary)
var VR{B};
var VC{B};

# current (real, imaginary)
var IR{L};
var IC{L};

# power injected on line at bus (real, imaginary)
var SR{L};
var SC{L};

# power generation (real, imaginary)
var SgenR{b in B, g in G[b]} >= SLR[b,g], <= SUR[b,g];
var SgenC{b in B, g in G[b]} >= SLC[b,g], <= SUC[b,g];

# auxiliary variables: V2 = |V|^2
var V2{b in B} >= VL[b]^2, <= VU[b]^2;

## objective function

# generation cost (WARNING: to remove quadratic power terms set C[g,2]=0)
minimize generation_costs: sum{b in B, g in G[b]} (C[b,g,2]*SgenR[b,g]^2 + C[b,g,1]*SgenR[b,g] + C[b,g,0]);

# # reactive power loss
# minimize reactive_power_loss: sum{b in B, g in G[b]} (C[b,g,1]*SgenC[b,g] + C[b,g,0]);

## constraints

# power flow (real, imaginary)
subject to power_flowR{b in B}:
   SDR[b] + sum{(b,a,i) in L} SR[b,a,i] = -shR[b]*V2[b] + sum{g in G[b]} SgenR[b,g];
subject to power_flowC{b in B}:
   SDC[b] + sum{(b,a,i) in L} SC[b,a,i] =  shC[b]*V2[b] + sum{g in G[b]} SgenC[b,g];

# definition of power in function of current and voltage (real, imaginary)
subject to power_injR{(b,a,i) in L}: SR[b,a,i] = VR[b]*IR[b,a,i] + VC[b]*IC[b,a,i];
subject to power_injC{(b,a,i) in L}: SC[b,a,i] = VC[b]*IR[b,a,i] - VR[b]*IC[b,a,i];

# Ohm's law ((b,a):real,imaginary; (a,b):real,imaginary)
subject to ohm1R{(b,a,i) in L0}:
  IR[b,a,i] = YbbR[b,a,i]*VR[b]-YbbC[b,a,i]*VC[b] + YbaR[b,a,i]*VR[a]-YbaC[b,a,i]*VC[a];
subject to ohm1C{(b,a,i) in L0}:
  IC[b,a,i] = YbbR[b,a,i]*VC[b]+YbbC[b,a,i]*VR[b] + YbaR[b,a,i]*VC[a]+YbaC[b,a,i]*VR[a];
subject to ohm2R{(b,a,i) in L0}:
  IR[a,b,i] = YabR[b,a,i]*VR[b]-YabC[b,a,i]*VC[b] + YaaR[b,a,i]*VR[a]-YaaC[b,a,i]*VC[a];
subject to ohm2C{(b,a,i) in L0}:
  IC[a,b,i] = YabR[b,a,i]*VC[b]+YabC[b,a,i]*VR[b] + YaaR[b,a,i]*VC[a]+YaaC[b,a,i]*VR[a];

# power bound on lines b->a defined on I
subject to power_bound_to{(b,a,i) in L : SU[b,a,i]>0 and SU[b,a,i]<Inf}:
  SR[b,a,i]^2 + SC[b,a,i]^2 <= SU[b,a,i]^2;

# definition of V2
subject to V2_def{b in B}: V2[b] = VR[b]^2 + VC[b]^2;

# bounds on phase difference
subject to phasediff1{(b,a,i) in L0}:
  VC[b]*VR[a] - VR[b]*VC[a] <= tan(pdUB[b,a,i]) * (VR[b]*VR[a] + VC[b]*VC[a]);
subject to phasediff2{(b,a,i) in L0}:
  VC[b]*VR[a] - VR[b]*VC[a] >= tan(pdLB[b,a,i]) * (VR[b]*VR[a] + VC[b]*VC[a]);
subject to phasediff3{(b,a,1) in L0}: VR[b]*VR[a] + VC[b]*VC[a] >= 0;

# reference bus: there had better be just one reference -- check in .run
subject to reference1{b in B : busType[b] == 3}: VC[b] = 0;
subject to reference2{b in B : busType[b] == 3}: VR[b] >= 0;

